import torch
from torch.utils.data import DataLoader, TensorDataset
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
import numpy as np
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt

# Set seed for reproducibility
np.random.seed(0)
torch.manual_seed(0)

# Cargamos los datos de "banana"
D = np.loadtxt( "bananaNorm.csv", delimiter=',', dtype=np.float32 )
print( D.shape )

# Dos características y dos clases
X = D[:,0:2]
y = D[:,2] - 1

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

train_dataset = TensorDataset(torch.tensor(X_train), torch.tensor(y_train))
test_dataset = TensorDataset(torch.tensor(X_test), torch.tensor(y_test))
# print( test_dataset.tensors[0].size(0) )
# print( test_dataset.tensors[0] )
# print( test_dataset.tensors )

train_loader = DataLoader(train_dataset, batch_size=16, shuffle=True)
test_loader = DataLoader(test_dataset, batch_size=16, shuffle=False)

# Define the neural network model
class SimpleNN(nn.Module):
	def __init__(self):
		super(SimpleNN, self).__init__()
		self.fc1 = nn.Linear(2, 8)
		self.fc2 = nn.Linear(8, 1)

	def forward(self, x):
		x = torch.relu( self.fc1(x) )
		x = torch.sigmoid( self.fc2(x) )
		return x

# Define the training function with regularization
def train_model(model, criterion, optimizer, train_loader, regularization_type=None, lambda_reg=0.01, epochs=20):
	epoch_losses = []
	
	for epoch in range(epochs):
		model.train()
		running_loss = 0.0
		
		for inputs, targets in train_loader:
			optimizer.zero_grad()
			outputs = model(inputs)
			loss = criterion(outputs.squeeze(), targets)
			
			# Apply L1 regularization
			if regularization_type == 'L1':
				l1_norm = sum(p.abs().sum() for p in model.parameters())
				loss += lambda_reg * l1_norm
			
			# Apply L2 regularization
			elif regularization_type == 'L2':
				l2_norm = sum(p.pow(2).sum() for p in model.parameters())
				loss += lambda_reg * l2_norm
			
			loss.backward()
			optimizer.step()
			
			running_loss += loss.item() * inputs.size(0)
		
		epoch_loss = running_loss / len(train_loader.dataset)
		epoch_losses.append(epoch_loss)
		print(f"Epoch {epoch+1}/{epochs}, Loss: {epoch_loss:.4f}")
	
	return epoch_losses

# Define the evaluation function
def evaluate_model(model, test_loader):
	model.eval()
	correct = 0
	total = 0
	
	with torch.no_grad():
		for inputs, targets in test_loader:
			outputs = model(inputs)
			predicted = (outputs.squeeze() > 0.5).float()
			total += targets.size(0)
			correct += (predicted == targets).sum().item()

	accuracy = correct / total
	print(f"Accuracy: {accuracy:.4f}")
	
	return accuracy

# Plot loss over epochs
def plot_training_loss(losses, title):
	plt.figure(figsize=(10, 5))
	plt.plot(range(1, len(losses) + 1), losses, marker='o')
	plt.title(title)
	plt.xlabel('Epochs')
	plt.ylabel('Loss')
	plt.grid(True)
	plt.show()

# Training and evaluating with L1 regularization
# print("Training with L1 Regularization:")
# model = SimpleNN()
# criterion = nn.BCEWithLogitsLoss()
criterion = nn.MSELoss()
# optimizer = optim.Adam(model.parameters(), lr=0.01)
# l1_losses = train_model(model, criterion, optimizer, train_loader, regularization_type='L1', lambda_reg=0.0, epochs=200)
# evaluate_model(model, test_loader)
# plot_training_loss(l1_losses, 'Training Loss with L1 Regularization')

# Reinitialize model, optimizer, and train with L2 regularization
print("\nTraining with L2 Regularization:")
model = SimpleNN()
optimizer = optim.Adam(model.parameters(), lr=0.007)
l2_losses = train_model(model, criterion, optimizer, train_loader, regularization_type='L2', lambda_reg=0.001, epochs=30)
evaluate_model(model, train_loader)

with torch.no_grad():
	outputs = model( train_dataset.tensors[0] )
	exactitud = accuracy_score( train_dataset.tensors[1], np.round(outputs) )
	print( "Exactitud entrenamiento=", exactitud )


evaluate_model(model, test_loader)
with torch.no_grad():
	outputs = model( test_dataset.tensors[0] )
	exactitud = accuracy_score( test_dataset.tensors[1], np.round(outputs) )
	print( "Exactitud prueba=", exactitud )

plot_training_loss(l2_losses, 'Training Loss with L2 Regularization')

